void RC_Update_And_Render(float32 _dt)
{
    RC_Read_Input(_dt);

	switch (RC_raycaster_state)
	{
        case RC_STATE_RAYCAST:
            RC_Raycast();
        break;

        case RC_STATE_RESET:
            RC_Reset();
            break;
	}
}

inline void RC_Raycast()
{
    // --- RAYCAST AND DRAW WALLS ---
    RC_Raycast_Walls();
    
    // --- RAYCAST AND DRAW FLOOR AND CEILING ---
    RC_Raycast_Floor_Ceiling();
}                                       

inline void RC_Raycast_Walls()
{
    // get local copy of buffer pointer - for 32 bpp
    u_int32* output_buffer_32 = RC_output_buffer_32;

    memset(output_buffer_32, 0, RC_prefs->screen_frame_buffer_size);

    // get local copy of RC_level_textures
    sBM_Bitmap_row* level_textures = RC_level_textures;

    // declare local cell structure for faster use
    sLV_Cell* level_map_cell = RC_level->map;

    // reset highest and lowest wall height
    RC_walls_start = 0;
    RC_walls_end = RC_render_height_i;

    // wall raycasting based on lodev.org algorithm
    for (u_int32 rx = 0; rx < RC_render_width_i; ++rx)
    {
        // normalized projection plane x coordinate: (-1.0, 1.0), precalculated - get from LUT
        float32 pp_coord_x = RC_pp_coord_x_LUT[rx];

        // new ray directions
        float32 ray_dx = RC_pp_dx + RC_pp_nsize_x_d2 * pp_coord_x; 
        float32 ray_dy = RC_pp_dy + RC_pp_nsize_y_d2 * pp_coord_x; 

        // coords of starting cell
        u_int8 curr_cell_x = (u_int8)RC_player_x;
        u_int8 curr_cell_y = (u_int8)RC_player_y;

        // length of ray from current position to next x or y-side
        float32 side_dist_x = 0.0f;
        float32 side_dist_y = 0.0f;

        // helpers for ray_dx and ray_dy
        float32 ray_dx_1d = 1.0f / ray_dx;
        float32 ray_dy_1d = 1.0f / ray_dy;

        // length of ray from one x or y-side to next x or y-side
        float32 delta_dist_x = fabsf(ray_dx_1d);
        float32 delta_dist_y = fabsf(ray_dy_1d);

        // what direction to step in x or y-direction (either +1 or -1)
        int8 step_x = 0;
        int8 step_y = 0;

        // was TOP-BOTTOM or LEFT-RIGHT hit ?
        u_int8 hit_side = 0;

        // what texture have we hit
        u_int8 wall_texture_index = 0;

        // preparing for DDA - calculate step and initial side dist
        if (ray_dx < 0.0f)
        {
            step_x = -1;
            side_dist_x = (RC_player_x - curr_cell_x) * delta_dist_x;
        }
        else
        {
            step_x = 1;
            side_dist_x = (curr_cell_x + 1.0f - RC_player_x) * delta_dist_x;
        }

        if (ray_dy < 0.0f)
        {
            step_y = -1;
            side_dist_y = (RC_player_y - curr_cell_y) * delta_dist_y;
        }
        else
        {
            step_y = 1;
            side_dist_y = (curr_cell_y + 1.0f - RC_player_y) * delta_dist_y;
        }

        // perform DDA
        u_int8 do_traverse = 1;

        while( do_traverse != 0 )
        {
            // jump to next map square, OR in x-direction, OR in y-direction
            if (side_dist_x < side_dist_y)
            {
                side_dist_x += delta_dist_x;
                curr_cell_x += step_x;
                hit_side = 0;
            }
            else
            {
                side_dist_y += delta_dist_y;
                curr_cell_y += step_y;
                hit_side = 1;
            }

            // get local copy of current cell for faster use
            sLV_Cell level_map_current_cell = level_map_cell[ curr_cell_x + (curr_cell_y << LV_MAP_SIZE_BITSHIFT) ];

            // check if ray has hit a solid wall
            if (level_map_current_cell.is_solid != 0)
            {
                // we must check the type of the cell
                switch (level_map_current_cell.cell_type)
                {
                    case LV_C_WALL_STANDARD:
                        wall_texture_index = level_map_current_cell.top_id;
                        do_traverse = 0;
                        break;
                }
            }
        }

        // calculate distance of perpendicular ray (Euclidean distance will give fisheye effect!)
        float32 wall_distance = 0.0f;

        if (hit_side == 0)  wall_distance = (curr_cell_x - RC_player_x + ((1 - step_x) >> 1)) * ray_dx_1d;
        else                wall_distance = (curr_cell_y - RC_player_y + ((1 - step_y) >> 1)) * ray_dy_1d;

        // converting wall_distance to intensity
        u_int32 intensity_index = (u_int32)(wall_distance * RC_INTENSITY_SCALE);
        intensity_index = MY_CLAMP(intensity_index, 0, RC_intensity_length_m1);

        // calculate height of line to draw on screen
        float32 wall_distance_1d = 1.0f / wall_distance;
        u_int16 pp_wall_height = (u_int16)(RC_render_height_i * wall_distance_1d);
        u_int16 pp_wall_height_d2 = pp_wall_height >> 1;

        // player z position helper
        float32 player_z_div_wall_distance = RC_player_z * wall_distance_1d;

        // calculate lowest and highest pixel to fill in current stripe
        int16 precalculated_helper = RC_render_height_d2_i + (int16)(RC_player_pitch + player_z_div_wall_distance);

        int16 tex_y_start = precalculated_helper - pp_wall_height_d2;
        int16 pp_wall_start = tex_y_start;

        // clamp starting y texture from 0 to abs()
        tex_y_start = (tex_y_start > 0) ? 0 : abs(tex_y_start); 
        pp_wall_start = MY_MAX(pp_wall_start, 0);

        int16 pp_wall_end = precalculated_helper + pp_wall_height_d2;        
        pp_wall_end = MY_CLAMP(pp_wall_end, 0, RC_render_height_i);

        // get min and max wall height for optimized floor and ceiling drawing
        RC_walls_start = MY_MAX(pp_wall_start, RC_walls_start);
        RC_walls_end = MY_MIN(pp_wall_end, RC_walls_end);

        // texturing calculations
        u_int8 intensity_value;

        // where exactly the wall was hit (0.0-1.0)
        float32 wall_hit_x;

        if (hit_side == 0)
        {
            intensity_value = RC_intensity_shading_wx_LUT[intensity_index];
            wall_hit_x = RC_player_y + wall_distance * ray_dy;
        }
        else
        {
            intensity_value = RC_intensity_shading_wy_LUT[intensity_index];
            wall_hit_x = RC_player_x + wall_distance * ray_dx;
        }

        wall_hit_x -= (int32)wall_hit_x;

        // converting hit x into x coordinate on the texture
        u_int16 tex_x = (u_int16)(wall_hit_x * RC_texture_size_f);

        if (hit_side == 0 && ray_dx < 0.0f) tex_x = RC_texture_size_m1_i - tex_x;
        if (hit_side == 1 && ray_dy > 0.0f) tex_x = RC_texture_size_m1_i - tex_x;

        // get local copy of current texture row and stripe
        u_int32** current_texture_x_row = level_textures[wall_texture_index].row;
        u_int32* current_texture_x_stripe = current_texture_x_row[tex_x];

        // this is for performance to avoid multiplication in loop
        u_int32 pp_wall_start_in_buffer = rx + pp_wall_start * RC_render_width_i;
        u_int32 pp_wall_end_in_buffer = rx + pp_wall_end * RC_render_width_i;

        // drawing walls into buffer - consider selected 16, 24, 32 depth mode
        if (RC_is_32)
        {
            // get local copy of premultiplied intensity table for individual channel
            u_int32* intensity_premultipled_CH1 = RC_intensity_premultiplied_LUT_255_CH1[intensity_value];
            u_int32* intensity_premultipled_CH2 = RC_intensity_premultiplied_LUT_255_CH2[intensity_value];
            u_int32* intensity_premultipled_CH3 = RC_intensity_premultiplied_LUT_255_CH3[intensity_value];

            u_int32* tex_y_coord_height = RC_pp_wall_height_ytex_LUT[pp_wall_height];
            u_int32* tex_y_coord__ptr = tex_y_coord_height + tex_y_start;

            u_int32* output_buffer_32__ptr = output_buffer_32 + pp_wall_start_in_buffer;

           for (u_int32 ry = pp_wall_start_in_buffer; ry < pp_wall_end_in_buffer; ry += RC_render_width_i)
           {
               u_int32 tex_y_coord = *tex_y_coord__ptr;
               tex_y_coord__ptr++;

                u_int32 texture_pixel = current_texture_x_stripe[tex_y_coord];

                u_int32 output_pixel =    intensity_premultipled_CH1[texture_pixel >> 16  & 0x0ff] |
                                            intensity_premultipled_CH2[texture_pixel >> 8   & 0x0ff] |
                                            intensity_premultipled_CH3[texture_pixel        & 0x0ff];
                                            
                *output_buffer_32__ptr = output_pixel;
                output_buffer_32__ptr += RC_render_width_i;
           }    
        }
    }
}

inline void RC_Raycast_Floor_Ceiling()
{  
    // get local copy of RC_level_textures
    sBM_Bitmap_row* level_textures = RC_level_textures;

    u_int32** current_texture_x_row = level_textures[0].row;
    sLV_Cell* level_map_cell = RC_level->map;

    // ray direction for leftmost ray (x = 0) and rightmost ray (x = width)
    float32 r_dx0 = RC_pp_dx - RC_pp_nsize_x_d2;
    float32 r_dy0 = RC_pp_dy - RC_pp_nsize_y_d2;

    float32 r_dx1_m_dx0_div_width = (RC_pp_nsize_x_d2 + RC_pp_nsize_x_d2) * RC_render_width_1d_f;
    float32 r_dy1_m_dy0_div_width = (RC_pp_nsize_y_d2 + RC_pp_nsize_y_d2) * RC_render_width_1d_f;

    // casting floor and ceiling - horizontal line by line - from left to right
    for (u_int16 ry = 0; ry < RC_render_height_i; ++ry)
    {
        // dont go further if the current floor/ceil scanline line won't be visible
        if (ry >= RC_walls_start)
        {
            if (ry < RC_walls_end)
            {
                ry = RC_walls_end;
                continue;
            }
        }

        u_int32 ry_m_render_width = ry * RC_render_width_i;        
        u_int32* output_buffer_32__ptr = RC_output_buffer_32 + ry_m_render_width;

        u_int8 is_floor = ry > RC_render_height_d2_i + RC_player_pitch;

        // SPLIT SCANLINE RENDER To FLOOR / CEILING (get rid of additional condition checking)
        // check if now rendering floor or ceiling scanline 
            
            if (is_floor != 0)
            {
                // current ray y position compared to the center of the screen (the horizon)
                float32 ry_pos = ry - RC_render_height_d2_i - RC_player_pitch;

                // vertical position of projection plane, 0.5 is between floor and ceiling
                float32 pp_z = RC_render_height_d2_i + RC_player_z;

                float32 straight_distance_to_point = pp_z / ry_pos;

                  u_int16 intensity_index = (u_int16)(straight_distance_to_point * RC_INTENSITY_SCALE);
                intensity_index = MY_CLAMP(intensity_index, 0, RC_intensity_length_m1);

                    // --- this part is DIFFERENT - intensity for floor ---
                    u_int8 intensity_value = RC_intensity_shading_f_LUT[intensity_index];

                // get local copy of premultiplied intensity table for individual channel
                u_int32* intensity_premultipled_CH1 = RC_intensity_premultiplied_LUT_255_CH1[intensity_value];
                u_int32* intensity_premultipled_CH2 = RC_intensity_premultiplied_LUT_255_CH2[intensity_value];
                u_int32* intensity_premultipled_CH3 = RC_intensity_premultiplied_LUT_255_CH3[intensity_value];

                // calculate the real world step vector we have to add for each x (parallel to camera plane)
                // adding step by step avoids multiplications with a weight in the inner loop
                float32 floor_step_x = straight_distance_to_point * r_dx1_m_dx0_div_width;
                float32 floor_step_y = straight_distance_to_point * r_dy1_m_dy0_div_width;

                float32 floor_x = RC_player_x + straight_distance_to_point * r_dx0;
                float32 floor_y = RC_player_y + straight_distance_to_point * r_dy0;

                // convert that values to fixed point
                int32 floor_x__fp = (int32)(floor_x * 65536.0f);
                int32 floor_y__fp = (int32)(floor_y * 65536.0f);

                int32 floor_step_x__fp = (int32)(floor_step_x * 65536.0f);
                int32 floor_step_y__fp = (int32)(floor_step_y * 65536.0f);

                u_int32 curr_cell_index = 0, prev_cell_index = 0;
                u_int8 texture_index = 0;

                // --- ACTUAL DRAWING FROM LEFT TO RIGHT - FLOOR ---
                    for (u_int32 ry_x = 0; ry_x < RC_render_width_i; ++ry_x)
                    {
                        floor_x__fp += floor_step_x__fp;
                        floor_y__fp += floor_step_y__fp;

                        if (*output_buffer_32__ptr != 0)
                        {
                            output_buffer_32__ptr++;
                            continue;
                        }

                        // the cell coord is simply got from the integer parts of floorX and floorY
                        u_int32 curr_cell_x = (floor_x__fp & FP_INTEGER_MASK_16) >> 16;
                        u_int32 curr_cell_y = (floor_y__fp & FP_INTEGER_MASK_16) >> 16;

                        u_int32 curr_cell_x_fraction__fp = (floor_x__fp & FP_FRACTION_MASK_16) >> 8;
                        u_int32 curr_cell_y_fraction__fp = (floor_y__fp & FP_FRACTION_MASK_16) >> 8;

                        curr_cell_index = curr_cell_x + (curr_cell_y << LV_MAP_SIZE_BITSHIFT);

                        if (curr_cell_index != prev_cell_index)
                        {
                            prev_cell_index = curr_cell_index;
                            texture_index = level_map_cell[curr_cell_index].floor_id;
                            current_texture_x_row = level_textures[texture_index].row;
                        }

                        // get the texture coordinate from the fractional part
                        u_int32 texture_pixel_x = (u_int32)((128 * 256 * curr_cell_x_fraction__fp) >> 16);
                        u_int32 texture_pixel_y = (u_int32)((128 * 256 * curr_cell_y_fraction__fp) >> 16);

                        // get pixel coords in texture
                        u_int32 texture_current_pixel = current_texture_x_row[texture_pixel_x][texture_pixel_y];

                        *output_buffer_32__ptr =
                                                                intensity_premultipled_CH1[texture_current_pixel >> 16 & 0x0ff] |
                                                                intensity_premultipled_CH2[texture_current_pixel >> 8 & 0x0ff] |
                                                                intensity_premultipled_CH3[texture_current_pixel & 0x0ff];

                        output_buffer_32__ptr++;
                    }
            }
            else
            {
                // current ray y position compared to the center of the screen (the horizon)
                float32 ry_pos = RC_render_height_d2_i - ry + RC_player_pitch;

                // vertical position of projection plane, 0.5 is between floor and ceiling
                float32 pp_z = RC_render_height_d2_i - RC_player_z;

                float32 straight_distance_to_point = pp_z / ry_pos;

                u_int16 intensity_index = (u_int16)(straight_distance_to_point * RC_INTENSITY_SCALE);
                intensity_index = MY_CLAMP(intensity_index, 0, RC_intensity_length_m1);

                    // --- this part is DIFFERENT - intensity for ceiling ---
                    u_int8 intensity_value = RC_intensity_shading_c_LUT[intensity_index];

                // get local copy of premultiplied intensity table for individual channel
                u_int32* intensity_premultipled_CH1 = RC_intensity_premultiplied_LUT_255_CH1[intensity_value];
                u_int32* intensity_premultipled_CH2 = RC_intensity_premultiplied_LUT_255_CH2[intensity_value];
                u_int32* intensity_premultipled_CH3 = RC_intensity_premultiplied_LUT_255_CH3[intensity_value];

                // calculate the real world step vector we have to add for each x (parallel to camera plane)
                // adding step by step avoids multiplications with a weight in the inner loop
                float32 floor_step_x = straight_distance_to_point * r_dx1_m_dx0_div_width;
                float32 floor_step_y = straight_distance_to_point * r_dy1_m_dy0_div_width;

                float32 floor_x = RC_player_x + straight_distance_to_point * r_dx0;
                float32 floor_y = RC_player_y + straight_distance_to_point * r_dy0;

                // convert that values to fixed point
                int32 floor_x__fp = (int32)(floor_x * 65536.0f);
                int32 floor_y__fp = (int32)(floor_y * 65536.0f);

                int32 floor_step_x__fp = (int32)(floor_step_x * 65536.0f);
                int32 floor_step_y__fp = (int32)(floor_step_y * 65536.0f);

                u_int32 curr_cell_index = 0, prev_cell_index = 0;
                u_int8 texture_index = 0;

                // --- ACTUAL DRAWING FROM LEFT TO RIGHT - CEILING ---
                for (u_int32 ry_x = 0; ry_x < RC_render_width_i; ++ry_x)
                {
                    floor_x__fp += floor_step_x__fp;
                    floor_y__fp += floor_step_y__fp;

                    if (*output_buffer_32__ptr != 0)
                    {
                        output_buffer_32__ptr++;
                        continue;
                    }

                    // the cell coord is simply got from the integer parts of floorX and floorY
                    u_int32 curr_cell_x = (floor_x__fp & FP_INTEGER_MASK_16) >> 16;
                    u_int32 curr_cell_y = (floor_y__fp & FP_INTEGER_MASK_16) >> 16;

                    u_int32 curr_cell_x_fraction__fp = (floor_x__fp & FP_FRACTION_MASK_16) >> 8;
                    u_int32 curr_cell_y_fraction__fp = (floor_y__fp & FP_FRACTION_MASK_16) >> 8;

                    curr_cell_index = curr_cell_x + (curr_cell_y << LV_MAP_SIZE_BITSHIFT);

                    if (curr_cell_index != prev_cell_index)
                    {
                        prev_cell_index = curr_cell_index;
                        texture_index = level_map_cell[curr_cell_index].ceil_id;
                        current_texture_x_row = level_textures[texture_index].row;
                    }

                    // get the texture coordinate from the fractional part
                    u_int32 texture_pixel_x = (u_int32)((128 * 256 * curr_cell_x_fraction__fp) >> 16);
                    u_int32 texture_pixel_y = (u_int32)((128 * 256 * curr_cell_y_fraction__fp) >> 16);

                    // get pixel coords in texture
                    u_int32 texture_current_pixel = current_texture_x_row[texture_pixel_x][texture_pixel_y];

               
                    
                    *output_buffer_32__ptr =
                                                            intensity_premultipled_CH1[texture_current_pixel >> 16 & 0x0ff] |
                                                            intensity_premultipled_CH2[texture_current_pixel >> 8 & 0x0ff] |
                                                            intensity_premultipled_CH3[texture_current_pixel & 0x0ff];

                    output_buffer_32__ptr++;
                }
            } 
    }
}